#define _GNU_SOURCE

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <errno.h>

/* Global variables */
char *conf_port;
char *conf_ipv4;
char *conf_ipv6;
char *conf_log;
char *conf_msgdir;
char *conf_maxconn;
char *conf_user, *conf_group;
char *conf_log_user_pass;
char *conf_debug;
char *conf_daemon;

/* Internal variables */
static char error[512];


char *ghostpop3_conf_strerror(void)
{
	return error;
}

char *fix(char *s)
{
	int last;
	char *ret;

	if (s == NULL)
		return NULL;

	/* begin */
	ret = s;
	while ((*ret == ' ') || (*ret == '\t'))
		ret++;

	/* end */
	last = strlen(ret) - 1;
	while ((ret[last] == ' ') || (ret[last] == '\t') || (ret[last] == '\n') || (ret[last] == '\r')) {
		ret[last] = '\0';
		last--;
	}

	return ret;
}

static void change(char **to, char *from)
{
	if (*to != NULL)
		free(*to);
	*to = strdup(from);
}

int ghostpop3_conf_load(char *file)
{
	FILE *f;
	char *line, *l, *r, *eq;
	char buf[4096];
	int lineno;

	strcpy(error, "");

	/* set defaults */
	conf_port = strdup("110");
	conf_ipv4 = strdup("on");
	conf_ipv6 = strdup("on");
	conf_log = strdup("/var/log/ghostpop3.log");
	conf_msgdir = strdup("/var/ghostpop3");
	conf_maxconn = strdup("1024");
	conf_user = strdup("nobody");
	conf_group = strdup("nogroup");
	conf_log_user_pass = strdup("on");
	conf_debug = strdup("0");
	conf_daemon = strdup("on");

	f = fopen(file, "r");
	if (!f) {
		snprintf(error, sizeof(error),
			"Cannot load conf file. Using defaults. [%s]", strerror(errno));
		return -1;
	}

	lineno = 0;
	while (!feof(f)) {
		line = fgets(buf, sizeof(buf), f);
		if (!line)
			break;

		lineno++;

		if ((buf[0] == '#') || (buf[0] == '\n'))
			continue;

		eq = strchr(line, '=');
		if (!eq) {
			snprintf(error, sizeof(error),
				"Error parsing line %d! Missing value.", lineno);
			goto err;
		}

		*eq = '\0';
		l = buf;
		r = eq + 1;

		l = fix(l);
		r = fix(r);
		/*printf("l=[%s] r=[%s]\n", l, r);*/

		if (strcmp(l, "port") == 0) {
			change(&conf_port, r);
		} else if (strcmp(l, "ipv4") == 0) {
			change(&conf_ipv4, r);
		} else if (strcmp(l, "ipv6") == 0) {
			change(&conf_ipv6, r);
		} else if (strcmp(l, "log") == 0) {
			change(&conf_log, r);
		} else if (strcmp(l, "msgdir") == 0) {
			change(&conf_msgdir, r);
		} else if (strcmp(l, "maxconn") == 0) {
			change(&conf_maxconn, r);
		} else if (strcmp(l, "user") == 0) {
			change(&conf_user, r);
		} else if (strcmp(l, "group") == 0) {
			change(&conf_group, r);
		} else if (strcmp(l, "log_user_pass") == 0) {
			change(&conf_log_user_pass, r);
		} else if (strcmp(l, "debug") == 0) {
			change(&conf_debug, r);
		} else if (strcmp(l, "daemon") == 0) {
			change(&conf_daemon, r);
		} else {
			snprintf(error, sizeof(error),
				"Error parsing line %d! Token not recognized.", lineno);
			goto err;
		}
	}

	return 0;

	err:
	fclose(f);
	return -1;

	return 0;
}
