#define _GNU_SOURCE

#include <errno.h>
#include <string.h>
#include <strings.h>
#include <stdlib.h>
#include <unistd.h>
#include <sys/poll.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <stdio.h>
#include <fcntl.h>
#include <netinet/in.h>
#include <sys/time.h>
#include <sys/ioctl.h>
#include <time.h>
#include <arpa/inet.h>
#include <stdarg.h>

#include <Ncgiall.h>
#include <Conn.h>

/* Global variables */
static time_t		now = 0;
static unsigned short	debug = 0;

static FILE	*Logf = NULL;
static char	*log_file = "ghostpop3test.log";
static int	port = 110;

static char	*commands [] = { "USER aaa\r\n", "PASS mumumu\r\n", "STAT\r\n", "QUIT\r\n", NULL };

struct priv
{
	unsigned short	comm_index;
};


static void Log(unsigned short level, char *format, ...)
{
	va_list	ap;

	if (level > debug)
		return;

	va_start(ap, format);
	vfprintf(Logf, format, ap);
	va_end(ap);
}

static void ghostpop3_connected(Conn *C)
{
	struct priv	*p;

	Log(4, "%s(A connection was estabilished on slot %d)\n",
		__FUNCTION__, C->slot);

	C->private = malloc(sizeof(struct priv));
	if (C->private == NULL) {
		Log(0, "cannot alloc memory!\n");
		exit(1);
	}

	p = (struct priv *) C->private;
	p->comm_index = 0;
}

static void ghostpop3_close(Conn *C)
{
	Log(5, "%s(Slot %d will close)\n",
		__FUNCTION__, C->slot);
	free(C->private);
}

static unsigned int ghostpop3_data(Conn *C)
{
	struct priv	*p;
	char *dump;
	unsigned int	ret;

	if (debug >= 8) {
		dump = Conn_dump(C->ibuf + C->ibuf_head, C->ibuf_tail - C->ibuf_head);
		Log(0, "data: recv: %s\n", dump);
		free(dump);
	}

	ret = C->ibuf_tail - C->ibuf_head;

	p = (struct priv *) C->private;

	if (commands[p->comm_index] == NULL) {
		Conn_close(C);
	} else {
		if (debug >= 8) {
			dump = Conn_dump(commands[p->comm_index], strlen(commands[p->comm_index]));
			Log(0, "data: send: %s\n", dump);
			free(dump);
		}
		Conn_enqueue(C, commands[p->comm_index], strlen(commands[p->comm_index]));
		p->comm_index++;
	}

	return ret;
}

static void ghostpop3_error(Conn *C)
{
	Log(1, "%s Slot=%d C=%p [%s]\n",
		__FUNCTION__, C ? C->slot : -1, (void *) C, Conn_strerror());
}

int main(void)
{
	char *stats;
	int i, ret;
	struct timeval start, end;
	unsigned int cps, elap;
	unsigned int ipv4conns = 2500, ipv6conns = 0;

	Logf = fopen(log_file, "a+");
	if (!Logf) {
		fprintf(stderr, "Cannot open log file [%s] [%s]\n",
			log_file, strerror(errno));
		return 1;
	}
	if (debug > 0)
		setlinebuf(Logf);

	Log(0, "Starting...\n");
	Log(0, "\tPort=%d\n", port);
	Log(0, "\tLogFile=%s Debug=%d\n", log_file, debug);

	Conn_debug(Logf, debug);

	ret = Conn_init(10000);
	if (ret == -1) {
		Log(0, "%s", Conn_strerror());
		return 1;
	}

	Conn_connected_cb = ghostpop3_connected;
	Conn_data_cb = ghostpop3_data;
	Conn_close_cb = ghostpop3_close;
	Conn_error_cb = ghostpop3_error;

	gettimeofday(&start, NULL);

	cps = 0;

	/* ipv4 */
	for (i = 0; i < ipv4conns; i++) {
		ret = Conn_connect(PF_INET, SOCK_STREAM, "127.0.0.1", 9110, 0, 0);
		if (ret == -1) {
			Log(0, "Error calling Conn_connect [%s]!\n",
				Conn_strerror());
			return 1;
		}
		cps++;
	}

	/* ipv6 */
	for (i = 0; i < ipv6conns; i++) {
		ret = Conn_connect(PF_INET6, SOCK_STREAM, "::1", 9110, 0, 0);
		if (ret == -1) {
			Log(0, "Error calling Conn_connect6 [%s]!\n",
				Conn_strerror());
			return 1;
		}
		cps++;
	}

	while (1) {
		ret = Conn_poll(1000);
		if (ret == -1) {
			Log(0, "Error in poll [%s]!\n",
				Conn_strerror());
			break;
		}

		if (ret == 0) {
			Log(0, "Finish work!\n");
			break;
		}

		if (debug >= 9) {
			stats = Conn_status(0);
			Log(9, "%s\n", stats);
			free(stats);
		}
	}

	gettimeofday(&end, NULL);

	elap = 1000000 * (end.tv_sec - start.tv_sec) + end.tv_usec - start.tv_usec;
	Log(0, "Finish in %ld.%06ld. ipv4conns=%d. ipv6conns=%d. Total=%d."
		" cps=%d. Conns per sec = %d\n",
		elap / 1000000, elap % 1000000,
		ipv4conns, ipv6conns, ipv4conns + ipv6conns,
		cps, cps / (elap / 1000000 + 1));

	return 0;
}
